;; Generated package description from auto-overlays.el
(define-package "auto-overlays" "0.10.8" "Automatic regexp-delimited overlays" 'nil :url "http://www.dr-qubit.org/emacs.php" :keywords '("extensions"))
